import {ipcMain, webContents} from 'electron'

/*
 * mqtt main端
 * @param {String} serverUrl mqtt服务地址
 * @param {Array} 中间件
 * */
class IpcMain implements ff.gc.Client {
  //渲染进程订阅的主题
  //number[]; 是webContentId数
  rendererTopics: {
    [topic: string]: number[]
  } = {}

  //主进程订阅的主题
  mainSubTopics: {
    [topic: string]: Function[]
  } = {}

  autoId: number = 1

  constructor() {
    //删除客户端
    ipcMain.on('$ipc/remove/client', (e: any, webContentId: number) => {
      //i 是主题
      for (let i in this.rendererTopics) {
        let index = this.rendererTopics[i].findIndex(id => webContentId === id)
        if (index > -1) {
          this.rendererTopics[i].splice(index, 1)

          if (this.rendererTopics[i].length === 0) {
            //当该主题已经没有id,直接取消订阅
            delete this.rendererTopics[i]
          }
        }
      }
    })

    /*
     * 发布一个ipc消息
     * @param {String} topic 主题
     * @param {mixni} message 内容
     * */
    ipcMain.on('$ipc/send', (e: any, topic: string, ...message: any) => {
      this.send(topic, ...message)
    })

    /*
     * 发布一个ipc消息
     * @param {String} topic 主题
     * @param {mixni} message 内容
     * */
    ipcMain.on('$ipc/sendmain', (e: any, topic: string, ...message: any) => {
      this.sendMain(topic, ...message)
    })

    //renderer主题订阅
    ipcMain.on('$ipc/subscribes', (e: any, tops: string[], webContentId: number) => {
      tops.forEach(topic => {
        if (this.rendererTopics[topic]) {
          if (!this.rendererTopics[topic].includes(webContentId)) this.rendererTopics[topic].push(webContentId)
        } else {
          this.rendererTopics[topic] = [webContentId]
        }
      })
    })

    //取消renderer主题订阅
    ipcMain.on('$ipc/unsubscribes', (e: any, tops: string[], webContentId: number) => {
      tops.forEach(topic => {
        if (!this.rendererTopics[topic]) return
        let index = this.rendererTopics[topic].findIndex(id => id === webContentId)
        if (index > -1) {
          this.rendererTopics[topic].splice(index, 1)
        }
        if (this.rendererTopics[topic].length === 0) {
          delete this.rendererTopics[topic]
        }
      })
    })
  }

  /*
   * 发布一个本地的mqtt,只会通知本机的订阅者
   * @param {String} topic 主题
   * @param {mixni} message 内容
   * */
  send(topic: string, message?: any, ...arg: any) {
    try {
      //浏览器添加时的特殊处理
      if ('标签页新增' === topic && message instanceof Array) {
        message.forEach(u => {
          this.autoId++
          u.id = this.autoId
        })
      }

      //获取订阅主题的数量,
      let clientIds: number[] = []

      //直接配置的，没有通配符
      if (this.rendererTopics[topic] && this.rendererTopics[topic].length) {
        clientIds = clientIds.concat(this.rendererTopics[topic])
      }

      if (!clientIds.length && !this.rendererTopics[topic] && !this.mainSubTopics[topic]) {
        ff.util.devlog('fgc 没有订阅该主题', topic)
        return
      }
      try {
        let str = typeof message !== 'object' ? String(message) : JSON.stringify(message)

        ff.util.devlog('fgc通知', topic, str.length > 50 ? str.substring(0, 50) + '...' : str) // message ?? '', ...arg
      } catch (e) {}

      clientIds = ff.util.arrayUnique(clientIds)

      clientIds.forEach(id => {
        let webc = webContents.fromId(id)
        webc && webc.send('$ipc/message', topic, message, ...arg)
      })

      //主程序主题
      this.sendMain(topic, message, ...arg)
    } catch (e) {
      console.error('错误了', e)
    }
  }

  /*
   * 发送到指定窗口
   * */
  sendToWebContents(webContactId: number, topic: string, message?: any) {
    //是否有订阅该主题
    //主题是否有包含webContactId
    if (this.rendererTopics[topic] && this.rendererTopics[topic].length && this.rendererTopics[topic].includes(webContactId)) {
      let webc = webContents.fromId(webContactId)
      webc && webc.send('$ipc/message', topic, message)
    }
  }

  sendMain(topic: string, message?: any, ...arg: any) {
    //主程序主题
    if (this.mainSubTopics[topic]) {
      this.mainSubTopics[topic].forEach(cb => cb(message, ...arg))
    }
  }

  sendCurrent(topic: string, ...message: any) {
    this.sendMain(topic, ...message)
  }

  //main中添加事件
  on<T extends string>(topic: T, param: T extends keyof ff.gc.Subscribes ? ff.gc.Subscribes[T] : (...p: any[]) => void) {
    if (this.mainSubTopics[topic]) {
      this.mainSubTopics[topic].push(param)
    } else {
      this.mainSubTopics[topic] = [param]
    }
  }

  //订阅
  subscribes(subs: Partial<ff.gc.Subscribes>): void {
    for (let i in subs) {
      let fun = subs[<keyof ff.gc.Subscribes>i]

      if (fun) this.on(i, fun)
    }
  }

  //取消订阅
  unsubscribes(subs: Partial<ff.gc.Subscribes>): void {
    for (let i in subs) {
      if (this.mainSubTopics[i]) {
        let topic = this.mainSubTopics[i]

        if (topic.length > 1) {
          topic.some((v, index) => {
            if (v === subs[<keyof ff.gc.Subscribes>i]) {
              topic.splice(index, 1)
              return true
            }
          })
        } else if (topic.length === 1 && topic[0] === subs[<keyof ff.gc.Subscribes>i]) {
          delete this.mainSubTopics[i]
        }
      }
    }
  }
}

let fgc = new IpcMain()

export default fgc
