import * as config from './linkedin/lisn.config'
/**
 * @用途 通讯类，linkedin、facebook等通讯
 * @author 黄敏
 * @创建时间 2022-08-11 10:26
 **/

interface CancelCb {
	(cancel: Axios.Canceler): void
}

export enum LinkedinAccount {
	nologin = 1, //未登录
	default, //普通账号
	sales //sales帐号
}

interface liErr {
	code: number
	msg: string
	time: number
	type: string
	conn_m: number
}

interface Stat {
	li_id: string | number
	li_uid?: string
	type: string
	code?: number
	msg?: string | object
}


let uidToUrnId: { [uid: string]: string } = {}


interface LinkedinUser {
	data: {
		plainId: number, //linkedin number id
		premiumSubscriber: boolean, //是否有买会员
	}
	included: {
		firstName: string
		lastName: string
		picture: {
			artifacts: { width: number, height: number, fileIdentifyingUrlPathSegment: string }[]
			rootUrl: string,
		}
		publicIdentifier: string, //linkedin uid
	}[]
}


interface CurrentUser extends Crawler.ContactAnchorFriends {
	is_pay: boolean
}

let baseUrl = 'https://www.linkedin.com/in/'

class Linkedin {
	readonly config = Object.freeze(config)
	readonly nologin = 1
	readonly default = 2
	readonly sales = 3

	login_status = this.nologin //登录状态，未登录，普通账号登录、lisn账号登录
	login_number_id = 0 //当前登录的linkedin数字ID
	current_user: CurrentUser | null = null

	getCompanyId(url: string) {
		let reg = /\/(?:company)\/([^\/]*?)(\/|\?|\S{0}$)/i.exec(url)
		if (reg) {
			return reg[1]
		}
	}

	getContactId(url: string) {
		let reg = /\/(?:in)\/([^\/]*?)(\/|\?|\S{0}$)/i.exec(url)
		if (reg) {
			return reg[1]
		}
	}


	//获取数据库的联系人
	async getDBContact(UID: string): Promise<Crawler.Contact | null> {
		//查详情本地数据库，是否有值
		let data = await ff.crawler.db.getContact(UID, 'linkedin-detail')
		if (data) return data

		//查基本本地数据库，是否有值
		data = await ff.crawler.db.getContact(UID, 'linkedin-base')
		return data
	}

	/**
	 * 获取linkedin的好友申请数量
	 * */
	async getInvitationCount() {
		let url = `https://www.linkedin.com/voyager/api/relationships/genericInvitationFacets?q=sent`
		await ff.util.asyncDelay(2000, 5000)

		let d: any = await ff.linkedin.ajax(url, {}, 'get')
		let p = d.elements.find((v: any) => v.invitationType === 'CONNECTION')
		return p ? p.count : 0
	}

	/**
	 * 获取linkedin的好友申请列表
	 * */
	async getInvitationList(page: number = 0, size = 10) {
		if (page < 0) return []
		let url = `https://www.linkedin.com/voyager/api/relationships/sentInvitationViewsV2?count=${size}&invitationType=CONNECTION&q=invitationType&start=${page}`
		let d: any = await ff.linkedin.ajax(url, {}, 'get')

		return d.elements
	}

	/*
	* 获取当前登录人的数据
	* */
	async getCurrentUser() {

		if (this.current_user) return this.current_user;

		let html = await ff.linkedin.ajax('https://www.linkedin.com/feed/', {}, 'get');
		// {"request":"/voyager/api/me","status":200,"body":"bpr-guid-14299"

		let reg = /\{\"request\"\:\"\/voyager\/api\/me\"\,\"status\":200\,"body":"(.+?)"/i.exec(html)

		if (!reg) throw ff.error('common_error', 'The data is wrong, maybe the linkedin api has changed')
		let id = reg[1]

		let d = ff.util.htmlToDom(html).querySelector<HTMLDivElement>(`#${id}`)

		if (!d) throw ff.error('common_error', 'The data is wrong, maybe the linkedin api has changed (2)')

		let obj: LinkedinUser = JSON.parse(d.innerText)

		if (!obj.included[0]) throw ff.error('common_error', 'The data is wrong, maybe the linkedin api has changed (3)')
		let user = obj.included[0];
		let img = ff.linkedin.extractImageUrl(user.picture)

		this.current_user = {
			origin_image: img,
			origin_name: ff.util.getFullName(user.firstName, user.lastName),
			origin_id: obj.data.plainId,
			info: baseUrl + user.publicIdentifier,
			is_pay: obj.data.premiumSubscriber
		}

		return this.current_user

	}


	/*
	 * 获取联系人基本信息
	 * @param {String} linkedin ID
	 * @returns {Object} baseInfo|null
	 * */
	async getContactBase(UID: string): Promise<Crawler.Contact | null> {
		//查详情本地数据库，是否有值
		let data = await this.getDBContact(UID)
		if (data) return data

		await ff.util.asyncDelay(1500, 4000)

		//无值，向下执行
		//网络请求
		let url = `https://www.linkedin.com/voyager/api/identity/dash/profiles?q=memberIdentity&memberIdentity=${UID}&decorationId=com.linkedin.voyager.dash.deco.identity.profile.FullProfileWithEntities-91`
		let code = await ff.linkedin.ajax(url)

		if (code === null || !code.elements || !code.elements.length) return null
		let codeActive = code.elements[0]

		//联系人姓名
		let firstName = ff.util.trim(codeActive.firstName)
		let lastName = ff.util.trim(codeActive.lastName)

		let model = ff.crawler.util.contactModel({
			origin: 'linkedin',
			plugin_id: 'linkedin-' + codeActive.publicIdentifier,
			completed: false,
			update_time: Date.now()
		})

		model.contacts_names.push({first_name: firstName ? firstName : ' ', last_name: lastName ? lastName : ' '})


		let friendInfo = await ff.linkedin.getFriendInfo(UID)

		let originAnchor: Crawler.Anchor = {
			type: 'linkedin',
			info: baseUrl + codeActive.publicIdentifier,
			friends: [],
			origin_id:this.getEndId(codeActive.objectUrn),
			premium: codeActive.premium ? 1 : 0,
			origin_name:ff.util.getFullName(firstName,lastName)
		}

		if (friendInfo.isFriend && friendInfo.user) {
			originAnchor.friends = [friendInfo.user]
		}

		let works: Crawler.Work[] = [],
				edus: Crawler.Education[] = [],
				icon = ''

		//工作经历
		let companyView = codeActive.profilePositionGroups
		if (companyView && companyView.elements && companyView.elements.length) {
			companyView.elements.forEach((companyWorks: any, i: number) => {
				let workGroup = (companyWorks.profilePositionInPositionGroup && companyWorks.profilePositionInPositionGroup.elements) || []

				if (!workGroup.length) return
				let thisWorks: Crawler.Work[] = []

				workGroup.forEach((work: any) => {
					let start_date, end_date, origin_id

					//开始时间
					if (work.dateRange && work.dateRange.start) {
						start_date = {
							year: work.dateRange.start.year,
							month: work.dateRange.start.month
						}
					}

					//结束时间
					if (work.dateRange && work.dateRange.end) {
						end_date = {
							year: work.dateRange.end.year,
							month: work.dateRange.end.month
						}
					}

					if (work.companyUrn) {
						//urn:li:fsd_company:2606072
						let id = this.getEndId(work.companyUrn)
						if (id) {
							origin_id = ff.crawler.util.createUniqueId(id, 'linkedin', 'company')
						}
					}

					let data: Crawler.Work = {
						location: work.geoLocationName || work.locationName,
						positions: work.title || '',
						company: work.companyName || '',
						start_date: start_date,
						end_date: end_date,
						descr: work.description || '',
						unique_id: origin_id,

						is_quit: false
					}

					//是否离职判断
					//有结束时间，为1已离职
					if (end_date) {
						data.is_quit = true
					} else if (start_date) {
						data.is_quit = false
					} else data.is_quit = i > 0

					if (data.company && data.company.length > 1) {
						thisWorks.push(data)
					}
				})

				works.push(...thisWorks)
			})
		}

		//教育经历
		let eduView = codeActive.profileEducations
		if (eduView && eduView.elements && eduView.elements.length) {
			eduView.elements.forEach((edu: any) => {
				let start_date, end_date

				//开始时间
				if (edu.dateRange && edu.dateRange.start) {
					start_date = {
						year: edu.dateRange.start.year,
						month: edu.dateRange.start.month || 0
					}
				}

				//结束时间
				if (edu.dateRange && edu.dateRange.end) {
					end_date = {
						year: edu.dateRange.end.year,
						month: edu.dateRange.end.month || 0
					}
				}

				let data = {
					college: edu.schoolName,
					degree_name: edu.degreeName,
					study: edu.fieldOfStudy,
					start_date: start_date,
					end_date: end_date
				}

				if (edu.schoolName && edu.schoolName.length > 1) {
					edus.push(data)
				}
			})
		}

		model.educations = edus

		if (codeActive.locationName) {
			model.geo_region = codeActive.locationName
		} else if (codeActive.geoLocation && codeActive.geoLocation.geo) {
			let geo = codeActive.geoLocation.geo

			if (geo.defaultLocalizedName) {
				model.geo_region = geo.defaultLocalizedName
			} else if (geo.country && geo.country.defaultLocalizedName) {
				model.geo_region = geo.country.defaultLocalizedName
			}
		}

		//国家
		let locationView = codeActive.primaryLocale
		let countryCode: any

		if (model.geo_region) {
			originAnchor.origin_address = model.geo_region
		}

		if (locationView.location && locationView.location.countryCode) {
			countryCode = locationView.location.countryCode.toUpperCase()
		} else if (locationView && locationView.country) {
			countryCode = locationView.country.toUpperCase()
		}

		if (countryCode) {
			model.country = countryCode
		} else {
			model.country = model.geo_region
		}

		model.anchors.push(originAnchor)

		//获取头像
		let pictureView = codeActive.profilePicture
		if (pictureView && codeActive.profilePicture && codeActive.profilePicture.displayImageReference) {
			let d = codeActive.profilePicture.displayImageReference
			if (d.vectorImage) {
				icon = this.extractImageUrl(d.vectorImage)
			}
		}

		model.imgs = icon ? [icon] : []

		//签名
		let signature = codeActive.headline
		model.remark = codeActive.summary || ''

		//work字段可能为空或work字段都是离职状态
		//查找签名是否有公司名或职位
		if (!works.length || works.every(v => v.is_quit)) {
			let newWork = {
				positions: '',
				company: '',
				descr: '',
				is_quit: false
			}

			if (signature) {
				let label = ff.crawler.util.labelTreatment(signature)

				let company = label[0]
				let title = label[1]

				newWork.company = company
				newWork.positions = title
				if (newWork.company && newWork.company.length > 1) {
					works.unshift(newWork)
				}
			}
		}

		model.works = works

		if (codeActive.birthDateOn) {
			if (codeActive.birthDateOn.year) model.birth_year = codeActive.birthDateOn.year
			if (codeActive.birthDateOn.month) model.birth_month = codeActive.birthDateOn.month
			if (codeActive.birthDateOn.day) model.birth_day = codeActive.birthDateOn.day
		}

		ff.crawler.db.saveContact(UID, model, 'linkedin-base')
		return model
	}





	get liErr() {
		return {
			get() {
				let liErr = localStorage.getItem('liErr') || '{"code":0,"msg":"","time":0,"type":""}'
				return JSON.parse(liErr)
			},
			set(liErr: liErr) {
				localStorage.setItem('liErr', JSON.stringify(liErr))
			},
			remove() {
				localStorage.removeItem('liErr')
			}
		}
	}

	/**
	 * 'https://www.linkedin.com/pub/vlad-cealicu/2/807/29a'
	 * linkedin 老格式转换，/pub/vlad-cealicu/2/807/29a
	 * */
	async pubToIn(url: string) {
		if (!/linkedin\.com\/pub\/.*/.test(url)) {
			return url
		}
		try {
			let e = await ff.axios(url)

			let uid = this.getContactId(e.request.responseURL)
			if (uid) {
				return baseUrl + uid
			} else {
				return ''
			}
		} catch (e) {
			ff.linkedin.ajaxErrorHandle(e)
			throw e
		}
	}


	/**
	 * 提取图片信息
	 * @param {{artifacts: any[]; rootUrl: string}} picture linkedin通用格式
	 * */
	extractImageUrl(picture: { artifacts: any[]; rootUrl: string }): string {
		let url = ''
		if (picture && picture.artifacts) {
			let imgs = picture.artifacts

			let maxWidth = 0
			let simg: any = null

			imgs.forEach(v => {
				if (v.width > maxWidth) {
					maxWidth = v.width
					simg = v
				}
			})

			let rootUrl = picture.rootUrl || ''
			let img = simg.fileIdentifyingUrlPathSegment || ''

			url = rootUrl + img
		}

		return url
	}


	//获取linkedin登录凭证
	async getToken(): Promise<string> {
		let token = await ff.crawler.util.getCookie('JSESSIONID', 'linkedin.com')
		if (!token) {
			token = await ff.crawler.util.getCookie('JSESSIONID', 'www.linkedin.com')
		}
		if (!token) token = ''
		return ff.util.trim(token, '"')
	}

	/**
	 * 获取当前登录者 数字ID
	 * https://www.linkedin.com/psettings/policy/notices?types=COMMENTS_VISIBILITY
	 * */

	async getNumberId() {
		if (!this.login_number_id) {
			let url = 'https://www.linkedin.com/psettings/policy/notices?types=COMMENTS_VISIBILITY'
			let r = await this.ajax(url, {}, 'get')
			if (r && r.content) {
				let u = r.content.member,
						arr = u.split(':')
				this.login_number_id = Number(arr[arr.length - 1])
			}
		}

		return this.login_number_id
	}

	/**
	 * 获取linkedin的密文id
	 * @param {string} uid uid是连接中 https://www.linkedin.com/in/yourk-soboros-181b0a63 中的 yourk-soboros-181b0a63
	 * */
	async getUrnId(uid: string) {

		if (uidToUrnId[uid]) return uidToUrnId[uid]

		//获取用户联系方式
		let url = 'https://www.linkedin.com/voyager/api/identity/profiles/' + uid + '/profileContactInfo'
		//urn:li:fs_contactinfo:ACoAAA2DOHEBYUISXyhKvgYlzk0ko6Gsbj7PT7Y
		let d: { entityUrn: string } = await ff.linkedin.ajax(url);

		let urnid = d.entityUrn.replace('urn:li:fs_contactinfo:', '');

		uidToUrnId[uid] = urnid

		return urnid
	}

	ajaxErrorHandle(e: any) {
		if (e.response) {
			//没有登录
			if (e.response.status === 999 || e.response.status === 401) {
				console.warn('linkedin.ts')
				ff.gc.send('登录状态linkedin更新', 'nologin')
				throw ff.error('no_login')
			} else if (e.response.status === 403) {
				throw ff.error('no_access', 'no access: The linkedin may not exist, or the user may have restricted access')
			} else if (e.response.status === 404) {
				throw ff.error('not_found')
			} else if (e.response.status === 429) {
				let err_data = {code: e.response.status}
				if (e.response.data) err_data = {...err_data, ...e.response.data}
				throw ff.error('frequency_too_fast', undefined, err_data)
			} else if (e.response.status === 451) {
				throw ff.error('linkedin_vpn')
			}
		}
		if (e.message === 'Network Error') {
			//没有网络
			throw ff.error('network_error')
		}
	}

	//linkedin ajax请求
	//@param {String} url 请求地址
	//@param {Object} data 请求数据
	//@param {String} type 请求类型，get或post
	async ajax(url: string, data: { [key: string]: any } = {}, type: 'get' | 'post' = 'get', cancelCb?: CancelCb, headers?: { [key: string]: string | number }): Promise<any> {
		try {
			let token = await this.getToken()
			if (!token) throw ff.error('no_login')
			let fHead = headers ? headers : {}
			let option: Axios.RequestConfig = {
				timeout: 60000,
				method: type,
				url: url,
				withCredentials: true, //可跨域访问
				headers: {
					'X-Requested-With': 'XMLHttpRequest',
					'csrf-token': token,
					'x-restli-protocol-version': '2.0.0',
					...fHead
				},
				xsrfHeaderName: 'csrf-token', //  默认
				cancelToken: new ff.axios.CancelToken(function (cancel: Axios.Canceler) {
					if (cancelCb) {
						cancelCb(cancel)
					}
				})
			}
			if (type === 'post') {
				option.data = data
			} else {
				option.params = {...data, fms: '1'}
			}
			let res = await ff.axios(option)

			return res.data
		} catch (e: any) {
			this.ajaxErrorHandle(e)
			throw e
		}
	}

	/*
	 * 获取由冒号(:)分隔的字符串最后一段id
	 * 格式：urn:li:fs_miniProfile:ACoAAAqEjnEBNiMcjmpeCpYs09szr2LFNeFN8dk
	 * urn:li:fsd_company:10142069
	 * */
	getEndId(urn: string | undefined) {
		let id: string = ''
		if (typeof urn === 'string' && urn.indexOf(':') > 2) {
			let a = urn.split(':')
			id = a[a.length - 1]
		}
		return id
	}

	//检查是否是好友
	//@param uid string
	//@param noCache 不拿缓存数据
	//@return FriendInfo
	async getFriendInfo(uid: string, noCache?: boolean): Promise<Crawler.FriendInfo> {
		let user = await this.getCurrentUser()
		let friendInfo: Crawler.FriendInfo = {
			isFriend: false,
			isPending: false,
			encrypt_uid: '',
			user: user
		}

		if (!noCache) {
			let d = await ff.crawler.db.getValidationFriend(uid, 'linkedin', user.origin_id)
			if (d) return d
		}

		let url = `https://www.linkedin.com/voyager/api/identity/profiles/${uid}/profileActions?contextType=NON_SELF_PROFILE_VIEW&profileActionTypes=List(ACCEPT,BLOCK,BUILD_RESUME,CONNECT,DISCONNECT,FOLLOW,IGNORE,INVITATION_PENDING,PERSONALIZED_CONNECT,RECOMMEND,REPORT,REQUEST_RECOMMENDATION,SAVE_TO_PDF,SIGNUP,UNFOLLOW,VIEW_PROFILE_IN_RECRUITER,VIEW_PROFILE_IN_SALES_NAVIGATOR,APPRECIATE,COMPOSE_MESSAGE,ADD_TEAMMATE)`
		let data = await this.ajax(url)

		console.warn('验证好友接口', data)

		if (!data) {
			console.log('linkedin好友验证结果返回为空：可能验证出问题')
			return friendInfo
		}

		if (data.entityUrn) {
			friendInfo.encrypt_uid = this.getEndId(data.entityUrn)
		}

		let longKey = 'com.linkedin.voyager.identity.profile.actions.GenericProfileAction'
		let pending = 'com.linkedin.voyager.identity.profile.actions.InvitationPending'
		let connect = 'com.linkedin.voyager.identity.profile.actions.Connect'
		let disconnect = 'com.linkedin.voyager.identity.profile.actions.Disconnect'
		let pAction = data.primaryAction

		if (pAction && pAction.action) {
			if (pAction.action[longKey] && pAction.action[longKey].composeOption) {
				let compose = data.primaryAction.action[longKey].composeOption

				if (compose) {
					let type = compose.composeOptionType

					if (type === 'REPLY' || type === 'CONNECTION_MESSAGE') {
						friendInfo.isFriend = true
					}

					if (compose.composeNavigationContext && compose.composeNavigationContext.existingConversationUrn) {
						friendInfo.message_id = this.getEndId(compose.composeNavigationContext.existingConversationUrn)
					}
				}
			} else if (pAction.action[pending]) {
				friendInfo.isPending = true
			}
		}

		//overflowActions 用于显示操作按钮
		// com.linkedin.voyager.identity.profile.actions.InvitationPending 表示显示了pending按钮
		let oAction = data.overflowActions
		if (!friendInfo.isFriend && !friendInfo.isPending && oAction && oAction.length) {
			oAction.some((e: any) => {
				if (!e.action) return false
				if (e.action[pending]) {
					friendInfo.isPending = true
					friendInfo.isFriend = false
					return true
				} else if (e.action[connect]) {
					friendInfo.isPending = false
					friendInfo.isFriend = false
					return true
				} else if (e.action[disconnect]) {
					friendInfo.isPending = false
					friendInfo.isFriend = true
					return true
				}
			})
		}

		await ff.crawler.db.saveValidationFriend(uid, 'linkedin', friendInfo, user.origin_id)
		return friendInfo
	}







	/*
	* 获取connections的列表
	* 默认为第一页
	* */
	async getConnections(url?: string): Promise<Crawler.Result> {

		let currentUser = await this.getCurrentUser()
		let result: Crawler.Result = {
			contacts: [],
			companys: []
		}

		if (!url)
			url =
					'https://www.linkedin.com/voyager/api/relationships/dash/connections?count=40&decorationId=com.linkedin.voyager.dash.deco.web.mynetwork.ConnectionListWithProfile-5&q=search&sortType=RECENTLY_ADDED'

		let api = url.replace(/&start=\d+/, '')

		let data = await ff.linkedin.ajax(url)

		if (!data || !data.elements || !data.elements.length) return result

		let total = data.paging.total || data.paging.count

		let batch: Crawler.Paging = {
			title: 'linkedin connections',
			api: api,
			uid: api,
			method: 'ff.linkedin.getConnections',
			param_start: 0,
			total: total,
			param_step: data.paging.count,
			param_name: 'start',
			page_size: data.paging.count
		}

		result.paging = batch

		data.elements.forEach((c: any) => {
			let profile = c.connectedMemberResolutionResult
			if (!profile) return
			let firstName = ff.util.trim(profile.firstName)
			let lastName = ff.util.trim(profile.lastName)
			let name = ff.util.getFullName(firstName, lastName)

			let contact = ff.crawler.util.contactModel({
				origin: 'linkedin',
				update_time: Date.now(),
				// name: name,
				completed: false,
				plugin_id: 'linkedin-' + profile.publicIdentifier
			})

			contact.contacts_names.push({first_name: firstName, last_name: lastName})
			contact.anchors.push({
				type: 'linkedin',
				info: `https://www.linkedin.com/in/${profile.publicIdentifier}`,
				origin_id:this.getEndId(profile.objectUrn),
				origin_name:ff.util.getFullName(firstName,lastName),
				friends: [currentUser]
			})

			//工作经历
			if (profile.headline) {
				let newWork = {
					positions: '',
					company: '',
					descr: '',
					is_quit: false
				}

				let label = ff.crawler.util.labelTreatment(profile.headline)
				let company = label[0]
				let title = label[1]
				newWork.company = company || name
				newWork.positions = title

				if (newWork.company.length > 1) {
					contact.works.push(newWork)
				}
			}

			if (profile?.profilePicture?.displayImageReference?.vectorImage) {
				let url = this.extractImageUrl(profile.profilePicture.displayImageReference.vectorImage)
				if (url) contact.imgs.push(url)
			}

			result.contacts?.push(contact)
		})

		return result
	}

}


setTimeout(() => {
	//通知Linkedin 状态
	ff.gc.on('登录状态linkedin更新', () => {
		ff.linkedin.login_status = ff.linkedin.nologin;
		ff.linkedin.login_number_id = 0;
		ff.linkedin.current_user = null;
	})
}, 10)

export default Linkedin
